/*****************************************************************************/
/*  rename_dialog.c - rename dialog box and ftp routines                     */
/*  Copyright (C) 1998-1999 Brian Masney <masneyb@newwave.net>               */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "ftp.h"

static void dorenCB(GtkWidget *widget, gpointer data);
static struct ftp_window_data *wdata;
static struct ftp_file_data *curfle;
static GtkWidget *edit;

void rename_dialog(gpointer data) {
   char tempstr[MAXSTR];
   
   wdata = (struct ftp_window_data *) data;
   if(wdata->local == -1) {
      ftp_log(LOG_MISC, "Rename: Not connected to a remote site\n");
      return;
   }
   else if(wdata->numselected != 1) {
      ftp_log(LOG_MISC, "Rename: You must only have one item selected\n");
      return;
   }

   if((curfle = get_next_selected_filename(wdata->host->files)) == NULL) {
      ftp_log(LOG_MISC, "Internal gFTP Error: Could not find a selected file. This is probably a bug. Please email masneyb@newwave.net about it\n");
      return;
   }

   g_snprintf(tempstr, sizeof(tempstr), "What do you want to rename %s to?", curfle->file);
   tempstr[sizeof(tempstr)-1] = '\0';
   edit = MakeEditDialog("Rename", tempstr, 
      "Rename", dorenCB, NULL,
      "Cancel", NULL, NULL);
}
/*****************************************************************************/
static void dorenCB(GtkWidget *widget, gpointer data) {
   char edttext[MAXSTR];
   int success;

   strncpy(edttext, gtk_entry_get_text(GTK_ENTRY(edit)), sizeof(edttext));
   edttext[sizeof(edttext)-1] = '\0';
   if(*edttext == '\0') {
     ftp_log(LOG_MISC, "Rename: Operation canceled...you must enter a string\n");
     return;
   }
   if(wdata->local == 1) {
      success = rename(curfle->file, edttext) == 0;
      if(!success) {
         ftp_log(LOG_MISC, "Error: Could not rename %s to %s: %s\n", curfle->file, edttext, g_strerror(errno));
      }
      else {
         ftp_log(LOG_MISC, "Successfully renamed %s to %s\n", curfle->file, edttext);
      }
   }
   else success = ftp_rename(wdata, wdata->host, curfle->file, edttext);
   if(success) refresh((gpointer) wdata);
}
/*****************************************************************************/
int ftp_rename(struct ftp_window_data *wdata, struct ftp_host_data *hdata, char *from, char *to) {
   if(ftp_sendcommand(1, wdata, hdata, NULL, 0, "RNFR %s\r\n", from) != '3') return(0);
   if(ftp_sendcommand(1, wdata, hdata, NULL, 0, "RNTO %s\r\n", to) != '2') return(0);
   return(1);
}
/*****************************************************************************/
